
body
{
	background: white;
}

pre
{
	overflow: visible;
	overflow-X: hidden;
	font-size: 100%;
}

#search-box,
#header,
#navigation,
#lastupdate,
#tools,
#footernav,
#copyright,
#github-ribbon,
#top,
.page-contents
{
	display: none;
}

div#content
{
	background: white;
	border: 0;
	margin: 0;
	padding: 0;
	font-family: serif;
	font-size: 10pt;
}
