h1 code { font-size: 1em; }
h1 code a { color: inherit; }

#copyright p {
	display: inline;
}

pre a { text-decoration: none; }
pre a:hover { text-decoration: underline; }

h1 code a {
	text-decoration: none;
	color: inherit;
	border-bottom: 1px dashed #CCC;
}

.hyphenate code {
	hyphens: manual;
}

col.caption {
	min-width: 170pt;
}

#navigation #cssmenu a.package {
	background-image: none;
}

span.tableEntryAnnotation { color: #866;}

a.protected {
	font-style: italic;
	background-image: url(../images/ddox/protected.png);
	background-repeat: no-repeat;
	background-position: 0pt 2pt;
	padding-left: 16px;
}
a.package {
	font-style: italic;
	background-image: url(../images/ddox/package.png);
	background-repeat: no-repeat;
	padding-left: 16px;
}
a.private {
	font-style: italic;
	background-image: url(../images/ddox/private.png);
	background-repeat: no-repeat;
	padding-left: 16px;
}
a.inherited:after { content: url(../images/ddox/inherited.png); padding-left: 3pt; vertical-align: bottom; }

#symbolSearchPane label { color: #ddd; }

#symbolSearch { width: 112pt; }

#symbolSearchResults {
	background: #F5F5F5;
	border: 1px solid #CCC;
	font-size: small;
	list-style: none;
	margin: 0;
	margin-top: 5px;
	padding: 0.3em;
	position: absolute;
	right: -1px;
	top: 100%;
	width: 100%;
	z-index: 1000;
}

#symbolSearchResults li {
	background-repeat: no-repeat;
	background-position: left center;
	padding-left: 18px;
}

#top #symbolSearchResults li a {
	color: #B03931;
	height: auto;
	padding: 0;
}
#symbolSearchResults li a:hover
{
	background: transparent;
	color: #742620;
	text-decoration: underline;
}

#symbolSearchResults .deprecated a { color: gray; }
#symbolSearchResults .module { background-image: url(../images/ddox/module.png); }
#symbolSearchResults .functiondeclaration { background-image: url(../images/ddox/function.png); }
#symbolSearchResults .classdeclaration { background-image: url(../images/ddox/class.png); }
#symbolSearchResults .interfacedeclaration { background-image: url(../images/ddox/interface.png); }
#symbolSearchResults .structdeclaration { background-image: url(../images/ddox/struct.png); }
#symbolSearchResults .variabledeclaration { background-image: url(../images/ddox/variable.png); }
#symbolSearchResults .property { background-image: url(../images/ddox/property.png); }
#symbolSearchResults .enumdeclaration { background-image: url(../images/ddox/enum.png); }
#symbolSearchResults .enummemberdeclaration { background-image: url(../images/ddox/enummember.png); }
#symbolSearchResults .aliasdeclaration { background-image: url(../images/ddox/alias.png); }
#symbolSearchResults .templatedeclaration { background-image: url(../images/ddox/template.png); }

/* Don't show simple handcrafted cheat sheets. DDOX does a good job generating
 * them. */
table.simple-cheatsheet { display: none; }
